"""
Scintillant -a tool for quickly creating skills adapted to work with the
Dialog service - the central system of the Lilia chat bot.
"""

GITHUB = "https://github.com"
BOTTLE_SKILL_TEMPLATE_URL = (
        GITHUB + "/PaperDevil/bottle-skill-template/archive/develop.zip"
)
AIOHTTP_SKILL_TEMPLATE_URL = ""
FAST_API_SKILL_TEMPLATE_URL = ""

# get key package details from scintillant/__version__.py
about = {
    '__title__': 'scintillant',
    '__description__': 'Fast bot creating framework',
    '__version__': '0.0.8',
    '__author__': 'Niel Ketov',
    '__author_email__': 'ketov-x@yandex.ru',
    '__url__': 'https://code.tatar.ru/projects/LILIYA/repos/scintillant',
    '__license__': 'Apache 2.0'
}  # type: ignore


"""
:mod: `client_models`
=====================

Description of models of client
applications interacting with requests.

.. module:: client_models
.. module_author:: Niel (Ketov) Gorev <ketov-x@ya.ru>
.. module_author:: http://t.me/ketov-x
"""
from dataclasses import dataclass, field
from http import HTTPStatus

from scintillant.apimodels.types import SkillResponseType, ClientResponseType


@dataclass
class ClientResponse(ClientResponseType):
    """Response model for the client application (Telegram, VK and others).

    .. note:: The skill does not return a response to the client,
    it is the task of the Dialog service.
    Inside the skill, we use SkillModels to receive requests from the Dialog
    and return the correct responses.

    Relationship diagram
    ---------------------
    Client (DialogAnswerRequest) -> Dialogs \n
    Dialogs (SkillRequest) -> Skill (SkillResponse) -> Dialogs \n
    Dialogs (**ClientResponse**) -> Client \n

    **parameters**, **types**
    -------------------------
    **types**::

    :type out_text: str.
    :type status_code: HTTPStatus.
    :type choices: list or None.
    :type client_app_action: str or None.
    :type requested_field_code: str or None.

    **parameters**::

    :param out_text: The text that the client will receive to respond to the
    user. May be HTML text.
    :param status_code: Internal status of the response to the client. [1]_
    :param choices: A set of lists displayed to the client in the form of
    buttons or lists. [2]_
    :param client_app_action: An action that can be performed by any client,
    if the interface allows it. [3]_
    :param requested_field_code: The name of the validator. [4]_

    **footprints**
    --------------

    [1] The specification of all statuses is currently under development,
    but here are a couple of values that clients are considering:
        - 200, Successful query execution;
        - 205, Resetting content, returning the bot to its original state;
        - 404, The class is not recognized or the skill responsible
        for the class is currently disabled;
        - 500, Internal error of the skill or Dialog.
    [2] The first item in the list is the visible value,
    the second is the returned payload.
    For instance: [["Choose the first", 1], ["Choose the second", 2]]
    [3] For example, delete the next message from a user (tek clients)
    or request a phone number (Exclusively telegrams).
    The list of actions available on clients is now expanding,
    but here are a couple of examples:
        - REQUEST_PHONE_NUMBER;
        - REMOVE_PREVIOUS_USER_MESSAGE.
    [4] This field is used at your own risk.
    it is supported by a very limited number of clients. For instance:
        - PHONE_NUMBER;
        - VEHICLE_NUMBER;
    """
    out_text: str = field()
    status_code: HTTPStatus or int = field(default=HTTPStatus.OK)
    choices: list or None = None
    client_app_action: str or None = None
    requested_field_code: str or None = None

    @classmethod
    def from_dict(cls, data: dict) -> ClientResponseType:
        """Generate ClientResponse from dict data.

        .. Example 1:: TODO
            >> import requests
            >> cr = ClientResponse._from_dict(requests.get(skill-url).json)
            >> print(cr)

        .. Example 2:: TODO
            >> cr = ClientResponse._from_dict({'out_text': ...})
            >> print(cr)

        :type data: dict

        :param data: Model Generation Data.
        :return: ClientResponse
        """
        return cls(
            out_text=data.get('out_text'),
            status_code=data.get('status_code', [500]),
            choices=data.get('choices'),
            client_app_action=data.get('client_app_action'),
            requested_field_code=data.get('requested_field_code')
        )

    @classmethod
    def from_skill_response(cls, response: SkillResponseType):
        """Generate ClientResponse from SkillResponse.

        Fastest way to get ClientResponse.
        For testing, you can use it right inside a skill,
        generating a skill response and matching against the
        ClientResponse class.

        .. note:: It is assumed that the method will be used after receiving
        a response from the skill to automatically generate
        a template response for the client.

        .. Example::
            >> import requests
            >> sr = SkillResponse._from_dict(requests.post(skill-url))
            >> cr = ClientResponse.from_skill_response(sr)

        :type response: SkillResponse

        :param response: HTTP response from the skill service.
        :return: ClientResponse
        """
        return cls.from_dict(response.serialize_json())

    def serialize_json(self) -> dict:
        """Returns a templated dictionary based on the fields of the class.

        This function is designed to send a class as JSON within HTTP requests
        or to respond to similar requests.

        .. Example aiohttp::
            >> @app.get('/')
            >> async def my_munc(request: Request):
            >>      cr = ClientResponse()
            >>      ...
            >>      return web.json_response(cr.serialize_json(), status=200)

        .. Example flask::
            >> @app.route('/', methods=['GET'])
            >> def my_munc():
            >>      cr = ClientResponse()
            >>      ...
            >>      return cr.serialize_json(), 200

        :return: dict
        """
        return {
            'out_text': self.out_text,
            'status_code': self.status_code,
            'choices': self.choices,
            'client_app_action': self.client_app_action,
            'requested_field_code': self.requested_field_code
        }



"""
:mod: `dialog_models`
=====================

Description of models of Dialog service
applications interacting with requests

.. module:: skill_models
.. module_author:: Niel (Ketov) Gorev <ketov-x@ya.ru>
.. module_author:: http://t.me/ketov-x
"""
from dataclasses import dataclass, field
from scintillant.apimodels.types import DialogAnswerRequestType


@dataclass()
class DialogAnswerRequest(DialogAnswerRequestType):
    """Request model for the Dialog service.

    Relationship diagram
    ---------------------
    Client (**DialogAnswerRequest**) -> Dialogs \n
    Dialogs (SkillRequest) -> Skill (SkillResponse) -> Dialogs \n
    Dialogs (ClientResponse) -> Client \n

    **parameters**, **types**
    -------------------------
    **types**::

    :type client_id: str.
    :type client_app: str.
    :type in_text: str or None.
    :type in_choice: str or None.

    **parameters**::

    :param client_id: str, Unique identifier for the customer, such as telegram ID or phone number. Not Null.
    :param client_app: str, The name of the client accessing the Dialog. [1]_
    :param in_text: str or None, The message that the user sent to the client. [2]_
    :param in_choice: User-made selection. [3]_

    **footprints**
    --------------

    [1] Clients are the services of the layer between the application servers and Dialog,
    but the name is correct for the application. Thus, for a bot responding to messages in telegrams,
    client_app will be equal to 'telegram'. Not Null.
    [2] This field is required if the in_choice field is empty.
    [3] The choice is made based on the values of the Choices field received in the ClientResponse.
    For instance:``
    >> choices = ClientResponse(...).choices # [["Choose the first", 1], ["Choose the second", 2]]
    >> da = DialogAnswerRequest(..., in_choice="Choose the first")``
    """
    client_id: str = field()
    client_app: str = field()
    in_text: str or None
    in_choice: str or None

    @classmethod
    def from_dict(cls, data: dict) -> DialogAnswerRequestType:
        """Generation DialogAnswerRequest from aiohttp.Request

        .. Example:: TODO
            >> @app.get('/')
            >> async def my_func(request: aiohttp.Request):
            >>      da = await DialogAnswerRequest.from_aio_request(request)
            >>      ...

        :type data: dict

        :param data: Model Generation Data.
        :return: DialogAnswerRequest
        """
        return cls(
            client_id=data.get('client_id'),
            client_app=data.get('client_app'),
            in_text=data.get('in_text'),
            in_choice=data.get('in_choice')
        )

    def serialize_json(self) -> dict:
        """Returns a templated dictionary based on the fields of the class.

        This function is designed to send a class as JSON within HTTP requests or to respond to similar requests.

        .. Example aiohttp::
            >> @app.get('/')
            >> async def my_munc(request: Request):
            >>      da = DialogAnswerRequest()
            >>      ...
            >>      return web.json_response(da.serialize_json(), status=200)

        .. Example flask::
            >> @app.route('/', methods=['GET'])
            >> def my_munc():
            >>      da = DialogAnswerRequest()
            >>      ...
            >>      return da.serialize_json(), 200

        :return: dict
        """
        return {
            'client_id': self.client_id,
            'client_app': self.client_app,
            'in_text': self.in_text,
            'in_choice': self.in_choice
        }


"""
:mod: `models`
=====================

Description of models within the skill network.

.. module:: skill_models
.. module_author:: Niel (Ketov) Gorev <ketov-x@ya.ru>
.. module_author:: http://t.me/ketov-x
"""

from scintillant.apimodels.models.client_models import ClientResponse
from scintillant.apimodels.models.skill_models import (
    SkillResponse, SkillRequest
)
from scintillant.apimodels.models.dialog_models import DialogAnswerRequest



"""
:mod: `skill_models`
=====================

Description of models of Skills
applications interacting with requests.

.. module:: skill_models
.. module_author:: Niel (Ketov) Gorev <ketov-x@ya.ru>
.. module_author:: http://t.me/ketov-x
"""
from dataclasses import dataclass, field
from datetime import datetime

from scintillant.apimodels.types import SkillResponseType, SkillRequestType


@dataclass
class SkillResponse(SkillResponseType):
    """The skill response to a Dialog service request.

    Relationship diagram
    ---------------------
    Client (DialogAnswerRequest) -> Dialogs \n
    Dialogs (SkillRequest) -> Skill (**SkillResponse**) -> Dialogs \n
    Dialogs (ClientResponse) -> Client \n

    **parameters**, **types**
    -------------------------
    **types**::

    :type status: str.
    :type out_text: str or None.
    :type choices: list or None.
    :type attach: None.
    :type client_app_action: str or None.
    :type request_field_code: str or None.
    :type context: dict or None.
    :type redirect_to: str or None.

    **parameters**::

    :param status: Responsible for the action that the Dialog will take after receiving a response. Not Null. [1]_
    :param out_text: The text that the client will receive to respond to the user. May be HTML text.
    :param choices: A set of lists displayed to the client in the form of buttons or lists. [2]_
    :param attach: This functionality is under development. [3]_
    :param client_app_action: An action that can be performed by any client, if the interface allows it. [4]_
    :param request_field_code: The name of the validator. [5]_
    :param context: Dictionary containing arbitrary JSON-serializable parameters. [6]_
    :param redirect_to: Pointer to the name of the skill to redirect the request to. [7]_

    .. note:: The context returned by the Dialog service can be used instead of the local data stores
    of the form Key: Value. This will allow the skills to save space and memory on a running machine,
    as well as respond more quickly to user requests.

    **footprints**
    --------------

    [1] For instance:
        - exit, Finishes the skill completely ignoring the out_text field.
        In this case, the user will receive a standard Dialog service start phrase in response.
        - wait,Having received this status, the Dialog service will give the user an answer that his request
        is being processed. After 3-5 seconds, the service will send a second request with the old parameters,
        expecting to receive a response. After the Dialog service receives the 'waiting' status for the 3rd time,
        the requests will stop going, and the user will receive an error message.
        - redirect, With this parameter, the Dialog service will ignore all other class fields except redirect_to.
        The user will receive a response from the skill whose name is written in redirect_to, if this skill exists
        and functions, otherwise the user will receive a response that the bot does not understand what the user wants.
    [2] The first item in the list is the visible value, the second is the returned payload.
    For instance: ``[["Choose the first", 1], ["Choose the second", 2]]``
    [3] In the future, it will allow sending images, audio files and other files to users.
    [4] For example, delete the next message from a user (tek clients) or request a phone number (Exclusively telegrams).
    The list of actions available on clients is now expanding, but here are a couple of examples:
        - REQUEST_PHONE_NUMBER;
        - REMOVE_PREVIOUS_USER_MESSAGE.
    [5] This field is used at your own risk. it is supported by a very limited number of clients. For instance:
        - PHONE_NUMBER;
        - VEHICLE_NUMBER.
    [6] The specified values will be saved by the Dialog service and sent unchanged in the all requests.
    [7] Without the status parameter equal to 'redirect', this parameter is ignored.
    """
    status: str = field()
    out_text: str or None = None
    choices: list or None = None
    attach: None = None  # TODO: Отправка изображений, звуков, карт

    client_app_action: str or None = None
    request_field_code: str or None = None

    context: dict or None = field(default_factory=dict)
    redirect_to: str or None = None

    @classmethod
    def from_dict(cls, data: dict) -> SkillResponseType:
        """Generate SkillResponse from dict data.

        .. Example 1:: TODO
            >> import requests
            >> sr = SkillResponse._from_dict(requests.get(skill-url).json)
            >> print(sr)

        .. Example 2:: TODO
            >> sr = SkillResponse._from_dict({'status': ...})
            >> print(sr

        :type data: dict.

        :param data: Model Generation Data.
        :return: SkillResponse.
        """
        return cls(
            status=data.get('status'),
            out_text=data.get('out_text'),
            choices=data.get('choices'),
            attach=None,
            client_app_action=data.get('client_app_action'),
            request_field_code=data.get('request_field_code'),
            context=data.get('context'),
            redirect_to=data.get('redirect_to')
        )

    @classmethod
    def from_response(cls, response) -> SkillResponseType:
        """Generate SkillResponse from basic requests.Response.

        .. Example::
            >> import requests
            >> sr = SkillResponse.from_response(requests.get('skill-url'))
            >> print(sr)

        :type response: requests.Response.

        :param response: HTTP response from the skill service.
        :return: SkillResponse.
        """
        json = response.json()
        return cls.from_dict(json)

    def serialize_json(self) -> dict:
        """Returns a templated dictionary based on the fields of the class.

        This function is designed to send a class as JSON within HTTP requests or to respond to similar requests.

        .. Example aiohttp::
            >> @app.get('/')
            >> async def my_munc(request: Request):
            >>      sr = SkillResponse()
            >>      ...
            >>      return web.json_response(sr.serialize_json(), status=200)

        .. Example flask::
            >> @app.route('/', methods=['GET'])
            >> def my_munc():
            >>      sr = SkillResponse()
            >>      ...
            >>      return sr.serialize_json(), 200

        :return: dict
        """
        return {
            "status": self.status,
            "out_text": self.out_text,
            "choices": self.choices,
            "client_app_action": self.client_app_action,
            "request_field_code": self.request_field_code,
            "context": self.context,
            "redirect_to": self.redirect_to
        }


@dataclass
class SkillRequest(SkillRequestType):
    """Skill request from Dialog service.

    Relationship diagram
    ---------------------
    Client (DialogAnswerRequest) -> Dialogs \n
    Dialogs (**SkillRequest**) -> Skill (SkillResponse) -> Dialogs \n
    Dialogs (ClientResponse) -> Client \n

    **parameters**, **types**
    -------------------------
    **types**::

    :type user: SkillRequest.User.
    :type update: SkillRequest.Update.
    :type context: dict or None

    **parameters**::

    :param user: Data about the user who sent the request. Not Null.
    :param update: Information of the request sent by the user. Not Null.
    :param context: Dictionary containing arbitrary JSON-serializable parameters. [1]_

    .. note:: Initially, the request for this model goes to a URL like ... /skill,
    but the application of this model is not limited to this route.
    This model can be used to write complex tests for skills in order to simulate
    a real request from the Dialog service.

    **footprints**
    --------------

    [1] The specified values will be saved by the Dialog service and sent unchanged in the all requests.
    """
    @dataclass
    class User:
        """Data about the user who sent the request.

        :type user_id: str.
        :type client_app: str.
        :type global_id: str or None.

        :param user_id: Unique identifier for the customer, such as telegram ID or phone number. Not Null.
        :param client_app: The name of the client accessing the Dialog.
        :param global_id: User identifier within the Dialog service system. [1]_

        [1] Using this identifier, you can request additional data about the user:
            - Phone number;
            - Email;
            - Authorizations.
        """
        user_id: str = field()
        client_app: str = field()
        global_id: str or None

    @dataclass
    class Update:
        """Information of the request sent by the user.

        :type in_text: str.
        :type in_choice: str.
        :type datetime: int.

        :param in_text: The message that the user sent to the client.
        This field is required if the in_choice field is empty.
        :param in_choice: User-made selection. [1]_
        :param datetime: Time of sending a request from a client service (Telegram, VK, etc.).

        [1] The choice is made based on the values of the Choices field received in the ClientResponse. For instance:``
        >> choices = ClientResponse(...).choices # [["Choose the first", 1], ["Choose the second", 2]]
        >> da = DialogAnswerRequest(..., in_choice="Choose the first")``
        """
        in_text: str
        in_choice: str
        datetime: int

    user: User
    update: Update
    context: dict

    @classmethod
    def from_dict(cls, data: dict) -> SkillRequestType:
        """Generate SkillRequest from dict data.

        .. Example:: TODO
            >> sr = SkillRequest._from_dict({'status': ...})
            >> print(sr

        :type data: dict.

        :param data: Model Generation Data.
        :return: SkillRequest.
        """
        user, upd, ctx = data.get('User'), data.get('Update'), data.get('Context')
        return cls(
            user=cls.User(
                user_id=user.get('user_id'),
                client_app=user.get('client_app'),
                global_id=user.get('global_id')
            ),
            update=cls.Update(
                in_text=upd.get('in_text'),
                in_choice=upd.get('in_choice'),
                datetime=upd.get('datetime', datetime.now().timestamp())
            ),
            context=ctx
        )

    def serialize_json(self) -> dict:
        """Returns a templated dictionary based on the fields of the class.

        This function is designed to send a class as JSON within HTTP requests or to respond to similar requests.

        .. Example aiohttp::
            >> @app.get('/')
            >> async def my_munc(request: Request):
            >>      sr = SkillRequest()
            >>      ...
            >>      return web.json_response(sr.serialize_json(), status=200)

        .. Example flask::
            >> @app.route('/', methods=['GET'])
            >> def my_munc():
            >>      sr = SkillRequest()
            >>      ...
            >>      return sr.serialize_json(), 200

        :return: dict
        """
        return {
            'User': self.user.__dict__,
            'Update': self.update.__dict__,
            'Context': self.context
        }



from http import HTTPStatus


class SkillResponseType:
    status: str
    out_text: str or None
    choices: list or None
    attach: None
    client_app_action: str or None
    request_field_code: str or None
    context: dict or None
    redirect_to: str or None

    @classmethod
    async def from_json_response(cls, response: dict): ...

    @classmethod
    async def from_request_response(cls, response): ...

    def serialize_json(self) -> dict: ...


class SkillRequestType:
    class User:
        user_id: str
        client_app: str
        global_id: str

    class Update:
        in_text: str
        in_choice: str
        datetime: int

    user: User
    update: Update
    context: dict

    def serialize_json(self) -> dict: ...

    @classmethod
    async def from_aio_request(cls, request): ...


class ClientResponseType:
    out_text: str
    status_code: HTTPStatus or int
    choices: list or None
    client_app_action: str or None
    requested_field_code: str or None

    @classmethod
    def from_dict(cls, data: dict): ...

    @classmethod
    def from_skill_response(cls, response: SkillResponseType): ...

    def serialize_json(self) -> dict: ...


class DialogAnswerRequestType:
    client_id: str
    client_app: str
    in_text: str or None
    in_choice: str or None

    @classmethod
    async def from_aio_request(cls, request): ...

    def serialize_json(self) -> dict: ...



"""
:mod: `postgres_db`
=====================

Methods for interacting with database systems that support the REST API.

.. module:: postgres_db
.. module_author:: Aminov Emil
.. module_author:: Niel (Ketov) Gorev <ketov-x@ya.ru>
.. module_author:: http://t.me/ketov-x
"""
import json

import requests
from requests.exceptions import HTTPError


class BaseDBClient:
    """Basic methods of interaction with the database.

    .. note:: The base client is a great way to write a convenient wrapper
    to get specific models from the database.
    Thus, you can create child classes that contain wrappers around the
    existing methods, or implement this class in another, for hidden access
    to the database.

    **parameters**, **types**
    -------------------------
    **types**::

    :type db_manager_url: str.

    **parameters**::

    :param db_manager_url: The address of the service providing the REST API
    for working with the database. [1]_

    **footprints**
    --------------

    [1] The URL specified this field must be the route to which the substring
    '/<model_name>/<primary_key>' is substituted and all of the following
    methods are executed: GET, POST, PUT, DELETE.
    """

    def __init__(self, db_manager_url: str):
        self.manager_url = db_manager_url + "/api/model"

    def get_object_by_id(self, model_name: str, pk: int or str):
        """Get model object by unique identifier (ID).

        :type model_name: str.
        :type pk: int or str.

        :param model_name: The name of the collection or table in the database.
        :param pk: The unique identifier of the received object.
        :return: dict.
        """
        request_path = f'{self.manager_url}/{model_name}/{pk}'

        def _get_object_by_id():
            """Executing a query on the model."""
            _data = requests.get(request_path)
            if not _data.status_code == 200:
                raise HTTPError(data['error_message'])
            return _data.json() if _data.status_code == 200 else None

        data = _get_object_by_id()
        response = data if data and len(data) > 0 else None
        return response

    def get_objects(self, model_name: str,
                    filters: dict = None,
                    limit: int = None,
                    offset: int = None):
        """Get a list of objects for a specific model.

        :type model_name: str.
        :type filters: dict.
        :type limit: int.
        :type offset: int.

        :param model_name: The name of the collection or table in the database.
        :param filters: They are indicated as a dictionary, the keys
        of which are possible fields (columns). [1]_
        :param limit: How many values to return. [2]_
        :param offset: How many values to ignore. [3]_
        :return: list.

        [1] Only those entities for which the statement is true will
        be returned: the column name is equal to one of the dictionary keys,
        and its value is equal to the corresponding value in the dictionary.
        [2] The maximum number of items specified will be returned. Those,
        if the base does not have the number of elements specified in the
        limit field, all available elements will be returned, otherwise None.
        [3] I mean how many received values from the beginning of the request
        should be skipped. Thus, if you specify offset equal to ten, then the
        first 10 lines of the query will be skipped.
        """
        filters = {} if not filters else filters
        request_path = f'{self.manager_url}/list/{model_name}'
        if limit and offset:
            response = requests.get(request_path + '?{}&{}&{}'.format(
                f'filters={json.dumps(filters)}',
                f'limit={limit}',
                f'offset={offset}'
            ))
        elif limit:
            response = requests.get(request_path + '?{}&{}'.format(
                f'filters={json.dumps(filters)}',
                f'limit={limit}'
            ))
        elif offset:
            response = requests.get(request_path + '?{}&{}'.format(
                f'filters={json.dumps(filters)}',
                f'offset={offset}'
            ))
        else:
            response = requests.get(request_path + '?{}'.format(
                f'filters={json.dumps(filters)}'
            ))
        return response.json()

    def insert_object(self, model_name: str, data: dict):
        """Create a new object of the corresponding model.

        :type model_name: str.
        :type data: dict.

        :param model_name: The name of the collection or table in the database.
        :param data: The data to be saved in the model.
        :return: dict.
        """
        request_path = f'{self.manager_url}/{model_name}'
        response = requests.post(request_path, json=data)
        data = response.json()
        if not response.status_code == 200:
            raise HTTPError(data['error_message'])
        return data

    def update_object(self, model_name: str, pk: int, data: dict):
        """Refresh data for a specific object in the model.

        :type model_name: str.
        :type pk: int.
        :type data: dict.

        :param model_name: The name of the collection or table in the database.
        :param data: The data to be saved in the model.
        :param pk: The unique identifier of the received object.
        :return: dict.
        """
        request_path = f'{self.manager_url}/{model_name}/{pk}'
        response = requests.patch(request_path, json=data)
        data = response.json()
        if not response.status_code == 200:
            raise HTTPError(data['error_message'])
        return data

    def put_object(self, model_name: str, pk: int, data: dict):
        """Replacing the data of a specific model object.

        :type model_name: str.
        :type pk: int.
        :type data: dict.

        :param model_name: The name of the collection or table in the database.
        :param data: The data to be saved in the model.
        :param pk: The unique identifier of the received object.
        :return: dict.
        """
        request_path = f'{self.manager_url}/{model_name}/{pk}'
        response = requests.put(request_path, json=data)
        data = response.json()
        if not response.status_code == 200:
            raise HTTPError(data['error_message'])
        return data

    def delete_object_by_id(self, model_name: str, pk: int):
        """Removes a specific object from the model data.

        :type model_name: str.
        :type pk: int.

        :param model_name: The name of the collection or table in the database.
        :param pk: The unique identifier of the received object.
        :return: dict.
        """
        request_path = f'{self.manager_url}/{model_name}/{pk}'
        response = requests.delete(request_path)
        data = response.json()
        if not response.status_code == 200:
            raise HTTPError(data['error_message'])
        return data

    def get_object_by_field(self, model_name: str,
                            field_name: str, field_value: object):
        """Returns one or more objects for which the value of the field
        'field_name' is equal to the value of 'field_value'.

        :type model_name: str.
        :type field_name: str.
        :type field_value: any.

        :param model_name: The name of the collection or table in the database.
        :param field_name: The name of the field or column by which the
        selection is made.
        :param field_value: A specific value that will be sampled.
        :return: dict or list
        """
        data = self.get_objects(model_name, filters={field_name: field_value})
        obj = data if data and len(data) == 0 else data[0]
        return obj


class DBUserClient(BaseDBClient):
    """An extension to the BaseDBClient that adds work
    exclusively with the user model.

    .. note:: For arguments, types and a detailed description of the models,
    see the description of the BaseDBClient class.
    """
    def __init__(self, db_manager_url: str):
        super().__init__(db_manager_url)

    def is_user_exist(self, client_app, client_id):
        """Checks if a user with the specified ID
        exists inside the Dialog service system

        :type client_app: str.
        :type client_id: str.

        :param client_app: The client application through which the
        user is accessing.
        :param client_id: Use the identifier inside the client application.
        :return: bool
        """
        user = self.get_object_by_field('User', client_app + '_id', client_id)
        return True if user and 'phone_number' in user else False

    def get_user_by_client_id(self, client_app: str,
                              user_id: int):
        """

        :param client_app:
        :param user_id:
        :return: dict
        """
        data = self.get_object_by_field('User', client_app + '_id', user_id)
        return data
        # TODO: требует добавления

    def add_token_for_user(self, client_app: str, user_id: int,
                           token_type: str, token: str):
        """

        :param client_app:
        :param user_id:
        :param token_type:
        :param token:
        :return:
        """
        data = self.get_user_by_client_id(client_app, user_id)
        if data:
            data['tokens'][token_type] = token
            updated_data = self.update_object('User', data.get('_id'), data)
            return updated_data
        raise HTTPError(f'No user found with id = {user_id}')


class DBClient(DBUserClient):
    """Export"""



from typing import Callable
from scintillant.apimodels import SkillResponse
from scintillant.apimodels.types import SkillRequestType, SkillResponseType


class ContextUpdater:
    states = {}  # state_name: state_func
    exit_phrases = ['exit', 'выход', 'хватит', 'остановись']

    def __init__(self, data: SkillRequestType):
        # Default parameters for find state and complete it
        self.data = data
        self.context = data.context
        self.response = SkillResponse(status='ok')

    @property
    def next_state(self):
        return self.context['state']

    @next_state.setter
    def next_state(self, func: Callable):
        if func:
            self.context['state'] = func.__name__
        else:
            self.context.pop('state')

    def execute_state(self):
        # Termination of the skill if the user has asked for it
        if (self.data.update.in_text in self.exit_phrases
                or self.data.update.in_choice in self.exit_phrases):
            self.response.status = 'exit'
        # In case the user has just entered the skill,
        # we call the _initial_state_ function.
        elif 'state' not in self.context:
            self._initial_state_()
        # In case the user already has a state.
        else:
            state = self.context.get('state')
            if state not in self.states:
                raise Exception(f"State {state} not registered!")
            else:
                self.states[state](self)

    def get_response(self) -> SkillResponseType:
        self.execute_state()
        self.response.context = self.context
        return self.response

    def _initial_state_(self):
        """Execute when state is None"""
        self.response.out_text = "Добро пожаловать в фреймворк Scintillant." \
                                 "\n\n" \
                                 "Вы сгенерировали ваш первый шаблон навыка " \
                                 "и видите это сообщение потому, что не " \
                                 "изменили ответ для шага `_initial_state`!"

    @classmethod
    def statefunc(cls, func):
        cls.states[func.__name__] = func

        def wrapper(*args, **kwargs):
            func(*args, **kwargs)
        wrapper.__name__ = func.__name__

        return wrapper



from setuptools import setup
from scintillant import about

# load the README file and use it as the long_description for PyPI
with open('README.md', 'r') as f:
    readme = f.read()

# package configuration - for reference see:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#id9
setup(
    name=about['__title__'],
    description=about['__description__'],
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    version=about['__version__'],
    author=about['__author__'],
    author_email=about['__author_email__'],
    url=about['__url__'],
    packages=[
        'scintillant',
        'scintillant.apimodels',
        'scintillant.apimodels.models',
        'scintillant.apimodels.db',
        'scintillant.controllers'
    ],
    include_package_data=True,
    python_requires=">=3.8.*",
    install_requires=['tqdm', 'requests', 'GitPython'],
    license=about['__license__'],
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'scintillant = scintillant.entry_points:main',
            'snlt = scintillant.entry_points:main'
        ],
    },
    keywords='package development template'
)



"""
scintillant.entry_points.py
~~~~~~~~~~~~~~~~~~~~~~

This module contains the entry-point functions for the scintillant module,
that are referenced in setup.py.
"""

from sys import argv, version_info
from os import getcwd, remove, rename, listdir, path
from shutil import move, rmtree
from zipfile import ZipFile

import requests

from scintillant import BOTTLE_SKILL_TEMPLATE_URL
from scintillant import AIOHTTP_SKILL_TEMPLATE_URL
from scintillant import FAST_API_SKILL_TEMPLATE_URL

from scintillant import about


def main() -> None:
    """Main package entry point.

    Delegates to other functions based on user input.
    """

    try:
        user_cmd = argv[1]
        if user_cmd in ['--bottle', 'bottle']:
            start_bottle_skill(skill_name=safe_list_get(argv, 2, None))
        elif user_cmd in ['--help', '-h', 'help']:
            print_help(safe_list_get(argv, 2, None))
        elif user_cmd in ['--version', '-v', 'version']:
            print(about['__version__'])
        else:
            raise RuntimeError(
                'please supply a command for scintillant - e.g. help.')
    except IndexError:
        raise RuntimeError('please supply a command for scintillant - e.g. help.')
    return None


def print_help(search_for: str = None):
    print("Scintillant -a tool for quickly creating skills adapted to work "
          "with the 'Dialog' service - the central system of the "
          "Lilia chat bot. \n\n"
          "** Commands ** \n"
          "+ bottle - Download the latest skill template based on the "
          "Bottle framework\n"
          ">> snlt bottle")


def start_bottle_skill(skill_name: str = None):
    """Download the latest skill template"""
    skill_name = input("Skill name: ") if not skill_name else skill_name
    skill_dirrectory = input(f"Skill directory ({skill_name}): ")
    if not skill_dirrectory:
        skill_dirrectory = skill_name
    print("Downloading project template...")

    r = requests.get(BOTTLE_SKILL_TEMPLATE_URL)
    with open('temp.zip', 'wb') as f:
        f.write(r.content)

    with ZipFile('temp.zip', 'r') as repo_zip:
        if skill_dirrectory == '.':
            repo_zip.extractall()
        else:
            repo_zip.extractall('.')

    remove('temp.zip')
    if not skill_dirrectory == '.':
        rename('bottle-skill-template-develop', skill_dirrectory)
    else:
        source = './bottle-skill-template-develop'
        files = listdir(source)
        for file in files:
            move(source + f'/{file}', getcwd())
        rmtree('./bottle-skill-template-develop')

    if not version_info < (3, 9):
        print("Language version below the required by the template")





def start_aiohttp_skill(skill_name: str = None):
    """Download the latest skill template"""
    skill_name = input("Skill name: ") if not skill_name else skill_name
    print("Downloading project template...")
    Repo.clone_from(AIOHTTP_SKILL_TEMPLATE_URL, getcwd() + f'/{skill_name}')


def start_fast_api_skill(skill_name: str = None):
    """Download the latest skill template"""
    skill_name = input("Skill name: ") if not skill_name else skill_name
    print("Downloading project template...")
    Repo.clone_from(FAST_API_SKILL_TEMPLATE_URL, getcwd() + f'/{skill_name}')


def safe_list_get(lst, idx, default):
    """Get list item, otherwise return standard value"""
    try:
        return lst[idx]
    except IndexError:
        return default
